#############################################################################
# dvdshrink_plugin.tcl Copyright 2005 r0lZ
# PgcEdit plugin to help DVDShrink users.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "DVDShrink" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more running under Windows to use this plugin.
# Of course you need also DVDShrink 3.2 installed on your system to use this
# plugin.
#############################################################################
# Version History:
#
# v 1.1 (September 7, 2005)
# - First public release
#############################################################################

if {$::tcl_platform(platform) == "windows"} {

	if {! [winfo exists .mainmenu.plugins.dvdshrink]} {
		# Add Plugins -> DVDShrink submenu
		.mainmenu.plugins add cascade -label "DVDShrink" -menu .mainmenu.plugins.dvdshrink
		menu .mainmenu.plugins.dvdshrink
		.mainmenu.plugins.dvdshrink add command -label "DVDShrink Remote Buttons" -command {::plugins::dvdshrink::remote_buttons}
		.mainmenu.plugins.dvdshrink add command -label "Check Streams Remapping safety" -command {::plugins::dvdshrink::check_streams_remapping}
		if {![catch {package require registry}]} {
			.mainmenu.plugins.dvdshrink add separator
			.mainmenu.plugins.dvdshrink add command -label "Launch DVDShrink (with current DVD)" -command {::plugins::dvdshrink::launch_dvdshrink true}
			.mainmenu.plugins.dvdshrink add command -label "Launch DVDShrink (empty)" -command {::plugins::dvdshrink::launch_dvdshrink false}
		}
		.mainmenu.plugins.dvdshrink add separator
		.mainmenu.plugins.dvdshrink add command -label "About" -command {
				tk_messageBox  -title $::plugins::dvdshrink::plugintitle -icon info \
						-message "$::plugins::dvdshrink::plugintitle for PgcEdit 0.6.0 or more.\nVersion $::plugins::dvdshrink::version\nAuthor: r0lZ\nFreeware!"
			}
	}	
	
	namespace eval ::plugins {}
	namespace eval ::plugins::dvdshrink {
	
		set version "1.1"
		set plugintitle "DVDShrink plugin"
		
		#############################################################################
		## add title button to skip to the next title on a DVD reauthored with DVDShrink
		proc remote_buttons {} {
			if {! [info exists ::dvddir]} {
				bell
				return
			}
			if {!$::config(expert_mode)} {
				set txt "Navigation buttons for DVDShrink reauthored movies"
				set txt "$txt\n\nThis plugin work ONLY on DVDShrink reauthored DVDs!"
				set txt "$txt\n"
				set txt "$txt\nThe macro will modify the commands of (almost) every PGCs,"
				set txt "$txt\nand create the VMGM Title menu PGC, so that the user can"
				set txt "$txt\nuse the 'Title' button of the remote to skip to the next title."
				set txt "$txt\n"
				set txt "$txt\nThanks goes to 2COOL for the idea.  See this tread:"
				set txt "$txt\nhttp://forum.doom9.org/showthread.php?threadid=61238"
				set txt "$txt\n"
				set txt "$txt\nThe macro will also allow to use the 'Prev' and 'Next Chapter'"
				set txt "$txt\nbuttons to be used in respectively the first and last chapter"
				set txt "$txt\nof a title, and the 'Return' button to go back to the beginning"
				set txt "$txt\nof the current title, regardless of the current chapter."
				set txt "$txt\nThe last option allow to automatically loop back to the beginning"
				set txt "$txt\nof the DVD when the last title has finished playing."
				set txt "$txt\n"
				set txt "$txt\nWARNING: Only some basic checks are made to ensure that"
				set txt "$txt\nthe DVD is in the correct DVDShrink reauthored format."
				set txt "$txt\nUse this macro only with DVDShrink reauthored DVDs!"
				set txt "$txt\nALWAYS backup the DVD IFOs before using this macro!"
				set txt "$txt\n"
				set txt "$txt\nContinue ?"
				set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon question -message $txt \
						-parent . -type yesno -default no]
				if {$r == "no"} return
			}
		
			### init GUI defaults to reflect current state
			if {$::menus(0,numlus)==1 && $::menus(0,1,numpgcs)==1} {
				set ::macros::dvdshrinkbuttons_title 1
			} else {
				set ::macros::dvdshrinkbuttons_title 0
			}
		
			set ::macros::dvdshrinkbuttons_prev [get_int ::pgcs(1,1,header) 0x9E 2]
			if {$::macros::dvdshrinkbuttons_prev > 1} {set ::macros::dvdshrinkbuttons_prev 1}
		
			set ::macros::dvdshrinkbuttons_next [get_int ::pgcs(1,1,header) 0x9C 2]
			if {$::macros::dvdshrinkbuttons_next > 1} {set ::macros::dvdshrinkbuttons_next 1}
		
			set ::macros::dvdshrinkbuttons_return [get_int ::pgcs(1,1,header) 0xA0 2]
			if {$::macros::dvdshrinkbuttons_return > 1} {set ::macros::dvdshrinkbuttons_return 1}
		
			if {[lsearch [get_type_cmds post ::pgcs($::pgcs(numvts),1)] \
					{30 08 00 00 01 00 00 00}] >= 0} {	;# CallSS FP
				set ::macros::dvdshrinkbuttons_loop 1
			} else {
				set ::macros::dvdshrinkbuttons_loop 0
			}
		
			### GUI
			if {[winfo exists .dvdshrinkbuttons]} {destroy .dvdshrinkbuttons}
			set w .dvdshrinkbuttons
			toplevel $w -class Dialog
			wm withdraw $w
			wm title $w $::plugins::dvdshrink::plugintitle
			if {$::config(tk) != "aqua"} {wm transient $w .}
			wm protocol $w WM_DELETE_WINDOW {set ::macros::dvdshrinkbuttons_done 0}
			bind $w <Escape> "set ::macros::dvdshrinkbuttons_done 0"
			bind $w <Return> "set ::macros::dvdshrinkbuttons_done 1"
		
			set txt "This macro allow to use the remote's buttons to navigate in DVDs reauthored with DVDShrink."
			label $w.lab -text $txt
			pack  $w.lab -padx 2 -pady 4
		
			frame $w.f1 -bd 2 -relief groove
			pack  $w.f1 -padx 4 -pady 2 -fill x
			checkbutton $w.f1.b2 -text "When in chapter 1, allow the 'Prev Chapter' button to go back to the beginning of the current title?" \
					-variable ::macros::dvdshrinkbuttons_prev
			pack  $w.f1.b2 -anchor w
			checkbutton $w.f1.b3 -text "When in last chapter, allow the 'Next Chapter' button to skip to the next title?" \
					-variable ::macros::dvdshrinkbuttons_next
			pack  $w.f1.b3 -anchor w
			checkbutton $w.f1.b6 -text "Always allow the 'Return' (GoUp) button to go back to the beginning of the current title?" \
					-variable ::macros::dvdshrinkbuttons_return
			pack  $w.f1.b6 -anchor w
			checkbutton $w.f1.b1 -text "Always allow the 'Title Menu' button to skip to the next title?" \
					-variable ::macros::dvdshrinkbuttons_title
			pack  $w.f1.b1 -anchor w
			checkbutton $w.f1.b4 -text "At the end of the last title, automatically loop back to the first title?" \
					-variable ::macros::dvdshrinkbuttons_loop
			pack  $w.f1.b4 -anchor w
			button $w.f1.b5 -text "Check all" -width 10 -command {
				set ::macros::dvdshrinkbuttons_title  1
				set ::macros::dvdshrinkbuttons_return 1
				set ::macros::dvdshrinkbuttons_prev   1
				set ::macros::dvdshrinkbuttons_next   1
				set ::macros::dvdshrinkbuttons_loop   1
			}
			pack  $w.f1.b5 -anchor w -padx 4 -pady 4
		
		
			# buttons
			frame  $w.fb
			pack   $w.fb -padx 2 -pady 4 -anchor e
			button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::dvdshrinkbuttons_done 1}
			pack   $w.fb.ok -side left -padx 2
			button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::dvdshrinkbuttons_done 0}
			pack   $w.fb.cancel -side left -padx 2
		
			update
			wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
			wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
			set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
			set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
			if {$wx < 10} {set wx 10}
			if {$wy < 22} {set wy 22}
			wm geometry $w +$wx+$wy
			wm deiconify $w
			focus $w
			grab set $w
		
			tkwait variable ::macros::dvdshrinkbuttons_done
			grab release $w
			destroy $w
			if {!$::macros::dvdshrinkbuttons_done} return
		
			if {$::macros::dvdshrinkbuttons_title} {
				### create or replace the VMGM menu
				set r "yes"
				if {$::menus(0,numlus)>0} {
					set txt "WARNING: There is already a VMGM Menu!"
					set txt "$txt\nIf you continue, it will be REPLACED by a new one!"
					set txt "$txt\n"
					set txt "$txt\nContinue anyway ?"
					set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning -message $txt \
							-parent . -type yesno -default yes]
				}
				if {$r == "yes"} {
					set ::raw_sectors(0,VMGM_PGCI_UT) \
							[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]
		
					::readdvd::parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
					::main::init_tree_menu
		
					### for each VTS except last one, replace the pre command "Set gprm(0) = 0"
					### by "Set gprm(0) = <current VTS number>"
					for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
						set cmds [get_type_cmds pre ::pgcs($vts,1)]
						set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
						if {$idx == -1} {
							set cmds [linsert $cmds 0 "71 00 00 00 00 [format %02X $vts] 00 00"]
						} else {
							lset cmds $idx "71 00 00 00 00 [format %02X $vts] 00 00"
						}
						set_type_cmds pre ::pgcs($vts,1) $cmds
					}
				}
			} else {
				### delete the VMGM menu
				set r "yes"
				if {$::menus(0,numlus)>0} {
					set txt "WARNING: The entire VMGM Menu will be DELETED!"
					set txt "$txt\n"
					set txt "$txt\nContinue anyway ?"
					set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning -message $txt \
							-parent . -type yesno -default yes]
					if {$r == "no"} return
				}
				if {$r == "yes"} {
					set ::raw_sectors(0,VMGM_PGCI_UT) ""
					set ::menus(0,numlus) 0
					::main::init_tree_menu
		
					### for each VTS except last one, replace the pre command
					### "Set gprm(0) = <current VTS number>" by "Set gprm(0) = 0"
					for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
						set cmds [get_type_cmds pre ::pgcs($vts,1)]
						set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
						if {$idx != -1} {
							lset cmds $idx "71 00 00 00 00 00 00 00"
						}
						set_type_cmds pre ::pgcs($vts,1) $cmds
					}
				}
			}
		
			if {$::macros::dvdshrinkbuttons_loop} {
				### Last VTS: replace Exit by CallSS FP
				set vts $::pgcs(numvts)
				set cmds [get_type_cmds post ::pgcs($vts,1)]
				if {$cmds == ""} {
					set cmds {{30 08 00 00 01 00 00 00}}
				} else {
					set idx [lsearch $cmds {30 01 00 00 00 00 00 00}]	;# Exit
					if {$idx >= 0} {
						lset cmds $idx "30 08 00 00 01 00 00 00"	;# CallSS FP
					}
				}
				set_type_cmds post ::pgcs($vts,1) $cmds
			} else {
				### Last VTS: replace CallSS-FP by Exit
				set vts $::pgcs(numvts)
				set cmds [get_type_cmds post ::pgcs($vts,1)]
				if {$cmds == ""} {
					set cmds {{30 01 00 00 00 00 00 00}}
				} else {
					set idx [lsearch $cmds {30 08 00 00 01 00 00 00}]	;# CallSS FP
					if {$idx >= 0} {
						lset cmds $idx "30 01 00 00 00 00 00 00"	;# Exit
					}
				}
				set_type_cmds post ::pgcs($vts,1) $cmds
			}
		
			### Allow to use the Prev button, when pressed during Chapter 1) to go back to the
			### beginning of the title, and the Next Button to go directly to the end of the Title
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				if {$::macros::dvdshrinkbuttons_prev} {
					set_int ::pgcs($vts,1,header) 0x9E 1 2
				} else {
					set_int ::pgcs($vts,1,header) 0x9E 0 2
				}
				if {$::macros::dvdshrinkbuttons_next} {
					set_int ::pgcs($vts,1,header) 0x9C 1 2
				} else {
					set_int ::pgcs($vts,1,header) 0x9C 0 2
				}
				if {$::macros::dvdshrinkbuttons_return} {
					set_int ::pgcs($vts,1,header) 0xA0 1 2
				} else {
					set_int ::pgcs($vts,1,header) 0xA0 0 2
				}
			}
		
			### util done.  Refresh listbox.
			refresh_current_pgc
			::main::dvd_modified true false
		
			# if there is an old VMGM VOB, delete it or move it to backup
			::utils::delete_vob_with_confirm "VIDEO_TS.VOB" true
		}
		
		#############################################################################
		## Safety check for DVDShrink's Logical remapping of enabled streams option.
		## Stored in: HKEY_CURRENT_USER\Software\DVD Shrink\DVD Shrink 3.2\Preferences\LogicalStreamRemapping
		proc check_streams_remapping {} {
			if {! [info exists ::dvddir]} {
				bell
				return
			}
		
			set used false
		
			# Titles
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
					set used [search_setstn_gprm ::pgcs($vts,$ttn)]
					if {$used} break
				}
				if {$used} break
			}
		
			# Menus
			if {!$used} {
				for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
					for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
						for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
							set used [search_setstn_gprm ::menus($vts,$lu,$pgc)]
							if {$used} break
						}
						if {$used} break
					}
					if {$used} break
				}
			}
		
			# Menu buttons
			if {!$used} {
				for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
					set butcmds {}
					if {[info exists ::buttons::butsinfo(menu,$vts,errorcode)] && \
							$::buttons::butsinfo(menu,$vts,errorcode) == 0} {
						for {set vid 1} {$vid <= $::buttons::butsinfo(menu,$vts,numvobids)} {incr vid} {
							for {set cid 1} {$cid <= $::buttons::butsinfo(menu,$vts,$vid,numcellids)} {incr cid} {
								if {[info exists ::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)] && \
											$::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts) > 0} {
									for {set group 1} {$group <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numgroups)} {incr group} {
										for {set but 1} {$but <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)} {incr but} {
											lappend butcmds $::buttons::butsinfo(menu,$vts,$vid,$cid,$group,$but,command)
										}
									}
								}
							}
						}
					}
					set used [search_setstn_gprm butcmds]
					if {$used} break
				}
				unset butcmds
			}
		
			# BOVs
			if {!$used} {
				for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
					set butcmds {}
					if {[info exists ::buttons::butsinfo(pgc,$vts,errorcode)] && \
							$::buttons::butsinfo(pgc,$vts,errorcode) == 0} {
						for {set vid 1} {$vid <= $::buttons::butsinfo(pgc,$vts,numvobids)} {incr vid} {
							for {set cid 1} {$cid <= $::buttons::butsinfo(pgc,$vts,$vid,numcellids)} {incr cid} {
								if {[info exists ::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)] && \
											$::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts) > 0} {
									for {set group 1} {$group <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numgroups)} {incr group} {
										for {set but 1} {$but <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)} {incr but} {
											lappend butcmds $::buttons::butsinfo(pgc,$vts,$vid,$cid,$group,$but,command)
										}
									}
								}
							}
						}
					}
					set used [search_setstn_gprm butcmds]
					if {$used} break
				}
				unset butcmds
			}
		
			set txt "Check DVDShrink Streams Remapping\n\n"
			set txt "${txt}\nThis macro checks if you may safely use the DVDShrink option\n"
			set txt "${txt}\"Logical remapping of enabled streams\" with the current DVD.\n\n"
			set txt "${txt}When this option is enabled, DVDShrink physically removes the\n"
			set txt "${txt}definitions of the stripped streams in the IFOs, and attempts to\n"
			set txt "${txt}change all SetSTN commands to point to the new string numbers.\n"
			set txt "${txt}However, it is unable to change the indirect SetSTN commands\n"
			set txt "${txt}(when a GPRM is used instead of a constant value), because the\n"
			set txt "${txt}number of the stream is not hardcoded in the command.\n"
			set txt "${txt}It is best to turn this option ON whenever possible, or there will be\n"
			set txt "${txt}references to non existing streams, which may confuse the player.\n"
			set txt "${txt}This means that it is better to use the logical remapping in Full Disc\n"
			set txt "${txt}mode but only if there are no indirect SetSTN command in the DVD.\n"
			set txt "${txt}Notes:\n"
			set txt "${txt}- This option is useless if you don't strip any stream.\n"
			set txt "${txt}- You may safely use it if you strip only the last streams (no gaps).\n"
			set txt "${txt}- The streams remapping option is ignored in Re-author mode.\n\n"
			set txt "${txt}Analysis results:\n\n"
			if {$used} {
				set txt "${txt}There are indirect SetSTN commands in this DVD.\n"
				set txt "${txt}(If you want to find them, search for \"\[41\" in the commands)\n"
				set txt "${txt}The logical remapping of enabled streams option should be OFF in\nFull Disc mode!"
				show_msg "DVDShrink's Logical remapping of enabled streams option is NOT safe."
				set rec "OFF"
				set default "no"
			} else {
				set txt "${txt}No indirect SetSTN commands found in this DVD.\n"
				set txt "${txt}The logical remapping of enabled streams option may safely be ON."
				show_msg "DVDShrink's Logical remapping of enabled streams option is safe."
				set rec "ON"
				set default "yes"
			}
			set txt "${txt}\n\nSummary for current DVD:\n\n"
			set txt "${txt}Recommended setting in Full Disc mode:  $rec  ($default)\n"
		
			set change false
			if {![catch {package require registry}]} {
				if {![catch {set cur [registry get \
						"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2\\Preferences" \
						"LogicalStreamRemapping"]}]} {
					set change true
					if {$cur == 0} {
						set cur OFF
					} else {
						set cur ON
					}
				}
			}
			if {$change} {
				set txt "${txt}Current DVDShrink 3.2's remapping option:  $cur.\n\n"
				set txt "${txt}Allow Logical remapping of enabled streams in DVDShrink 3.2 ?"
				if {$default == "yes" && $cur == "ON"} {set default "cancel"}
				if {$default == "no" && $cur == "OFF"} {set default "cancel"}
				set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon info \
						-message $txt -parent . -type yesnocancel -default $default]
				if {$r == "yes"} {
					catch {registry set \
						"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2\\Preferences" \
						"LogicalStreamRemapping" 1}
				} elseif {$r == "no"} {
					catch {registry set \
						"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2\\Preferences" \
						"LogicalStreamRemapping" 0}
				}
			} else {
				tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon info -message $txt -parent .
			}
			return $used
		}
		proc search_setstn_gprm {cmds_upvar} {
			upvar $cmds_upvar cmds
			if {[string first "\{41 " $cmds] == -1} {
				return false
			} else {
				return true
			}
		}
		
		#############################################################################
		## Launch DVDShrink, with current DVD (if there is one DVD loaded).
		proc launch_dvdshrink {with_current_dvd} {
			if {![catch {package require registry}]} {
				if {![catch {set filename [registry get \
						"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2\\Preferences" \
						"InstallApp"]}]} {
					if {[info exists ::dvddir] && $with_current_dvd} {
						if {$::dvd_modified} {
							set r [tk_messageBox -title $::plugins::dvdshrink::plugintitle -icon warning \
									-message "DVD modified but not saved!\n\nThe DVD will be loaded in DVDShrink\nwith the files currently saved on disk.\n\nContinue?" \
									-parent . -type yesno -default no]
							if {$r == "no"} {
								return
							}
						}
						exec -- $filename $::dvddir
					} else {
						exec -- $filename
					}
				} else {
					bell
				}
			} else {
				bell
			}
		}
	} ;# end of namespace ::plugins::dvdshrink

} else {
	# not under Win platform
	tk_messageBox -title "DVDShrink plugin" -icon error \
			-message "This plugin works only under Windows.  Sorry."
} ; # end of if {$::tcl_platform(platform) == "windows"}
